import numpy as np
from eight_mile.utils import exporter, Offsets


__all__ = []
export = exporter(__all__)


@export
# Metrics UAS/LAS
# UCM (unlabeled complete matching rate, the % of sentences with whole correct trees
# LCM labeled complete matching rate
class LAS:
    def __init__(self):
        self.total = 0
        self.correct = 0

    def add(self, heads_pred, heads_gold, labels_pred, labels_gold):
        valid = labels_gold != Offsets.PAD
        for h_p, h_g, l_p, l_g in zip(heads_pred[valid], heads_gold[valid], labels_pred[valid], labels_gold[valid]):
            correct = h_p == h_g and l_p == l_g
            self.total += 1
            self.correct += float(correct)

    @property
    def name(self):
        return 'las'

    @property
    def score(self):
        return self.correct / self.total


@export
class UAS:
    def __init__(self):
        self.total = 0
        self.correct = 0

    def add(self, heads_pred, heads_gold, labels_pred, labels_gold):
        valid = labels_gold != Offsets.PAD
        for h_p, h_g in zip(heads_pred[valid], heads_gold[valid]):
            correct = h_p == h_g
            self.total += 1
            self.correct += float(correct)

    @property
    def name(self):
        return 'uas'

    @property
    def score(self):
        return self.correct / self.total


@export
class LCM:
    def __init__(self):
        self.total = 0
        self.correct = 0

    def add(self, heads_pred, heads_gold, labels_pred, labels_gold):
        valid = labels_gold != Offsets.PAD
        correct = True
        # For LCM, the entire tree has to match with labels
        for h_p, h_g, l_p, l_g in zip(heads_pred[valid], heads_gold[valid], labels_pred[valid], labels_gold[valid]):
            correct = correct and (h_p == h_g and l_p == l_g)

        self.total += 1
        self.correct += int(correct)

    @property
    def name(self):
        return 'ucm'

    @property
    def score(self):
        return self.correct / self.total


@export
class UCM:
    def __init__(self):
        self.total = 0
        self.correct = 0

    def add(self, heads_pred, heads_gold, labels_pred, labels_gold):
        valid = labels_gold != Offsets.PAD
        correct = True
        # For UCM, the entire tree has to match
        for h_p, h_g in zip(heads_pred[valid], heads_gold[valid]):
            correct = correct and (h_p == h_g)

        self.total += 1
        self.correct += int(correct)

    @property
    def name(self):
        return 'ucm'

    @property
    def score(self):
        return self.correct / self.total



import gzip
import hashlib
import logging
import os
import shutil
import zipfile
from six.moves.urllib.request import urlretrieve
from eight_mile.utils import (
    mime_type,
    exporter,
    read_json,
    write_json,
    validate_url,
)
from eight_mile.progress import create_progress_bar

logger = logging.getLogger('mead.layers')

__all__ = []
export = exporter(__all__)


def _verify_file(file_loc):
    # dropbox doesn't give 404 in case the file does not exist, produces an HTML. The actual files are never HTMLs.
    if not os.path.exists(file_loc):
        return False

    if os.path.isfile(file_loc) and mime_type(file_loc) == "text/html":
        return False

    return True


@export
def delete_old_copy(file_name):
    if os.path.exists(file_name):
        if os.path.isfile(file_name):
            os.remove(file_name)
        else:
            shutil.rmtree(file_name)
    return file_name


@export
def extract_gzip(file_loc):
    temp_file = delete_old_copy("{}.1".format(file_loc))
    with gzip.open(file_loc, 'rb') as f_in:
        with open(temp_file, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    if mime_type(temp_file) == "application/x-tar":
        return extract_tar(temp_file)
    else:
        shutil.move(temp_file, file_loc)
        return file_loc



@export
def extract_tar(file_loc):
    import tarfile
    temp_file = delete_old_copy("{}.1".format(file_loc))
    with tarfile.open(file_loc, "r") as tar_ref:
        tar_ref.extractall(temp_file)
    if len(os.listdir(temp_file)) != 1:
        raise RuntimeError("tar extraction unsuccessful")
    return os.path.join(temp_file, os.listdir(temp_file)[0])


@export
def extract_zip(file_loc):
    temp_file = delete_old_copy("{}.1".format(file_loc))
    with zipfile.ZipFile(file_loc, "r") as zip_ref:
        zip_ref.extractall(temp_file)
    return temp_file


@export
def extractor(filepath, cache_dir, extractor_func):
    with open(filepath, 'rb') as f:
        sha1 = hashlib.sha1(f.read()).hexdigest()
    logger.info("extracting file..")
    path_to_save = filepath if extractor_func is None else extractor_func(filepath)
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    path_to_save_sha1 = os.path.join(cache_dir, sha1)
    delete_old_copy(path_to_save_sha1)
    shutil.move(path_to_save, path_to_save_sha1)
    logger.info("downloaded data saved in {}".format(path_to_save_sha1))
    return path_to_save_sha1


@export
def get_file_or_url(file, cache=None):
    """
    Return itself if its a file, else downloaded.  If cache, use that
    :param file: The file name
    :param cache: A cache location, If `None`, just download
    :return: The proper filename
    """

    if validate_url(file):
        logger.info(f'Downloading {file}')
        if cache:
            file = SingleFileDownloader(file, cache).download()
        else:
            file, _ = urlretrieve(file)
    return file


@export
def open_file_or_url(file, mode='r', cache=None):
    """
    Return handle to a file, if a url, download first. If cache, use that
    :param file: The file name
    :param model: The open mode
    :param cache: A cache location, If `None`, just download
    :return: The handle
    """
    return open(get_file_or_url(file, cache), mode)


@export
def web_downloader(url, path_to_save=None):
    # Use a class to simulate the nonlocal keyword in 2.7
    class Context: pg = None

    def _report_hook(count, block_size, total_size):
        if Context.pg is None:
            length = int((total_size + block_size - 1) / float(block_size)) if total_size != -1 else 1
            Context.pg = create_progress_bar(length)
        Context.pg.update()

    if not path_to_save:
        path_to_save = "/tmp/data.dload-{}".format(os.getpid())
    try:
        path_to_save, _ = urlretrieve(url, path_to_save, reporthook=_report_hook)
        Context.pg.done()
    except Exception as e:  # this is too broad but there are too many exceptions to handle separately
        raise RuntimeError("failed to download data from [url]: {} [to]: {}".format(url, path_to_save))
    return path_to_save



@export
def is_file_correct(file_loc, data_dcache=None, key=None):
    """check if the file location mentioned in the json file is correct, i.e.,
    exists and not corrupted. This is needed when the direct download link/ path for a file
    changes and the user is unaware. This is not tracked by sha1 either. If it returns False, delete the corrupted file.
    Additionally, if the file location is a URL, i.e. exists in the cache, delete it so that it can be re-downloaded.

    Keyword arguments:
    file_loc -- location of the file
    data_dcache -- data download cache location (default None, for local system file paths)
    key -- URL for download (default None, for local system file paths)
    """
    if _verify_file(file_loc):
        return True
    # Some files are prefixes (the datasset.json has `train` and the data has `train.fr` and `train.en`)
    dir_name = os.path.dirname(file_loc)
    # When we are using this for checking embeddings file_loc is a url so we need this check.
    if os.path.exists(dir_name):
        files = [os.path.join(dir_name, f) for f in os.listdir(dir_name) if os.path.join(dir_name, f).startswith(file_loc)]
        if files and all(_verify_file(f) for f in files):
            return True
    delete_old_copy(file_loc)
    if key is not None:  # cache file validation
        update_cache(key, data_dcache)
    return False


@export
def is_dir_correct(dir_loc, dataset_desc, data_dcache, key, ignore_file_check=False):
    """check if the directory extracted from the zip location mentioned in the datasets json file is correct, i.e.,
    all files inside exist and are not corrupted. If not, we will update the cache try to re-download them.

    Keyword arguments:
    dir_loc -- location of the directory
    dataset_desc -- to know the individual file locations inside the directory
    data_dcache -- data download cache location
    key -- URL for download
    ignore_file_check --to handle enc_dec datasets, see later.
    """

    if not os.path.exists(dir_loc) or not os.path.isdir(dir_loc):
        update_cache(key, data_dcache)
        return False
    if ignore_file_check:  # for enc_dec tasks there's no direct downloads
        return True
    files = [os.path.join(dir_loc, dataset_desc[k]) for k in dataset_desc if k.endswith("_file")]
    for f in files:
        if not is_file_correct(f, key, data_dcache):
            return False
    return True


DATA_CACHE_CONF = "data-cache.json"


@export
def update_cache(key, data_download_cache):
    dcache = read_json(os.path.join(data_download_cache, DATA_CACHE_CONF))
    if key not in dcache:
        return
    del dcache[key]
    write_json(dcache, os.path.join(data_download_cache, DATA_CACHE_CONF))




@export
class Downloader:
    ZIPD = {'application/gzip': extract_gzip, 'application/zip': extract_zip}

    def __init__(self, data_download_cache, cache_ignore):
        super().__init__()
        self.cache_ignore = cache_ignore
        self.data_download_cache = data_download_cache

    def download(self):
        pass


@export
class SingleFileDownloader(Downloader):
    def __init__(self, dataset_file, data_download_cache, cache_ignore=False):
        super().__init__(data_download_cache, cache_ignore)
        self.dataset_file = dataset_file
        self.data_download_cache = data_download_cache

    def download(self):
        file_loc = self.dataset_file
        if is_file_correct(file_loc):
            return file_loc
        elif validate_url(file_loc):  # is it a web URL? check if exists in cache
            url = file_loc
            dcache_path = os.path.join(self.data_download_cache, DATA_CACHE_CONF)
            dcache = read_json(dcache_path)
            if url in dcache and is_file_correct(dcache[url], self.data_download_cache, url) and not self.cache_ignore:
                logger.info("file for {} found in cache, not downloading".format(url))
                return dcache[url]
            else:  # download the file in the cache, update the json
                cache_dir = self.data_download_cache
                logger.info("using {} as data/embeddings cache".format(cache_dir))
                temp_file = web_downloader(url)
                dload_file = extractor(filepath=temp_file, cache_dir=cache_dir,
                                       extractor_func=Downloader.ZIPD.get(mime_type(temp_file), None))
                dcache.update({url: dload_file})
                write_json(dcache, os.path.join(self.data_download_cache, DATA_CACHE_CONF))
                return dload_file
        raise RuntimeError("the file [{}] is not in cache and can not be downloaded".format(file_loc))


@export
class AddonDownloader(Downloader):
    ADDON_SUBPATH = 'addons'
    """Grab addons and write them to the download cache
    """
    def __init__(self, dataset_file, data_download_cache, cache_ignore=False):
        super().__init__(data_download_cache, cache_ignore)
        self.dataset_file = dataset_file
        self.data_download_cache = data_download_cache

    def download(self):
        file_loc = self.dataset_file
        if is_file_correct(file_loc):
            return file_loc
        elif validate_url(file_loc):  # is it a web URL? check if exists in cache
            url = file_loc
            dcache_path = os.path.join(self.data_download_cache, DATA_CACHE_CONF)
            dcache = read_json(dcache_path)
            # If the file already exists in the cache
            if url in dcache and is_file_correct(dcache[url], self.data_download_cache, url) and not self.cache_ignore:
                logger.info("file for {} found in cache, not downloading".format(url))
                return dcache[url]
            # Otherwise, we want it to be placed in ~/.bl-cache/addons
            else:  # download the file in the cache, update the json
                cache_dir = self.data_download_cache
                addon_path = os.path.join(cache_dir, AddonDownloader.ADDON_SUBPATH)
                if not os.path.exists(addon_path):
                    os.makedirs(addon_path)
                path_to_save = os.path.join(addon_path, os.path.basename(file_loc))
                logger.info("using {} as data/addons cache".format(cache_dir))
                web_downloader(url, path_to_save)
                dcache.update({url: path_to_save})
                write_json(dcache, os.path.join(self.data_download_cache, DATA_CACHE_CONF))
                return path_to_save
        raise RuntimeError("the file [{}] is not in cache and can not be downloaded".format(file_loc))

def _update_md(dataset_description_map, download_dir):
    updated = {k: v for k, v in dataset_description_map.items()}
    if download_dir:
        updated.update({k: os.path.join(download_dir, dataset_description_map[k]) for k in dataset_description_map.keys() if k.endswith('_file')})
    return updated

@export
class DataDownloader(Downloader):
    def __init__(self, dataset_desc, data_download_cache, enc_dec=False, cache_ignore=False):
        super().__init__(data_download_cache, cache_ignore)
        self.dataset_desc = dataset_desc
        self.data_download_cache = data_download_cache
        self.enc_dec = enc_dec

    def download(self):
        dload_bundle = self.dataset_desc.get("download", None)
        if dload_bundle is not None:  # download a zip/tar/tar.gz directory, look for train, dev test files inside that.
            dcache_path = os.path.join(self.data_download_cache, DATA_CACHE_CONF)
            dcache = read_json(dcache_path)
            if dload_bundle in dcache and \
                    is_dir_correct(dcache[dload_bundle], self.dataset_desc, self.data_download_cache, dload_bundle,
                                   self.enc_dec) and not self.cache_ignore:
                download_dir = dcache[dload_bundle]
                logger.info("files for {} found in cache, not downloading".format(dload_bundle))
                updated = _update_md(self.dataset_desc, download_dir)
                return updated
            else:  # try to download the bundle and unzip
                if not validate_url(dload_bundle):
                    raise RuntimeError("can not download from the given url")
                else:
                    cache_dir = self.data_download_cache
                    temp_file = web_downloader(dload_bundle)

                    download_dir = extractor(filepath=temp_file, cache_dir=cache_dir,
                                             extractor_func=Downloader.ZIPD.get(mime_type(temp_file), None))
                    if "sha1" in self.dataset_desc:
                        if os.path.split(download_dir)[-1] != self.dataset_desc["sha1"]:
                            raise RuntimeError("The sha1 of the downloaded file does not match with the provided one")
                    dcache.update({dload_bundle: download_dir})
                    write_json(dcache, os.path.join(self.data_download_cache, DATA_CACHE_CONF))
                    updated = _update_md(self.dataset_desc, download_dir)
                    return updated
        else:  # we have download links to every file or they exist
            updated = _update_md(self.dataset_desc, None)
            if not self.enc_dec:
                updated.update({k: SingleFileDownloader(self.dataset_desc[k], self.data_download_cache).download()
                        for k in self.dataset_desc if k.endswith("_file") and self.dataset_desc[k]})
            return updated


@export
class EmbeddingDownloader(Downloader):
    def __init__(self, embedding_file, embedding_dsz, embedding_sha1, data_download_cache, cache_ignore=False, unzip_file=True):
        super().__init__(data_download_cache, cache_ignore)
        self.embedding_file = embedding_file
        self.embedding_key = embedding_dsz
        self.data_download_cache = data_download_cache
        self.unzip_file = unzip_file
        self.sha1 = embedding_sha1

    @staticmethod
    def _get_embedding_file(loc, key):
        if os.path.isfile(loc):
            logger.info("embedding file location: {}".format(loc))
            return loc
        else:  # This is a directory, return the actual file
            files = [x for x in os.listdir(loc) if str(key) in x]
            if len(files) == 0:
                raise RuntimeError("No embedding file found for the given key [{}]".format(key))
            elif len(files) > 1:
                logger.info("multiple embedding files found for the given key [{}], choosing {}".format(key, files[0]))
            embed_file_loc = os.path.join(loc, files[0])
            return embed_file_loc

    def download(self):
        if is_file_correct(self.embedding_file):
            logger.info("embedding file location: {}".format(self.embedding_file))
            return self.embedding_file
        dcache_path = os.path.join(self.data_download_cache, DATA_CACHE_CONF)
        dcache = read_json(dcache_path)
        if self.embedding_file in dcache and not self.cache_ignore:
            download_loc = dcache[self.embedding_file]
            logger.info("files for {} found in cache".format(self.embedding_file))
            return self._get_embedding_file(download_loc, self.embedding_key)
        else:  # try to download the bundle and unzip
            url = self.embedding_file
            if not validate_url(url):
                raise RuntimeError("can not download from the given url")
            else:
                cache_dir = self.data_download_cache
                temp_file = web_downloader(url)
                unzip_fn = Downloader.ZIPD.get(mime_type(temp_file)) if self.unzip_file else None
                download_loc = extractor(filepath=temp_file, cache_dir=cache_dir,
                                         extractor_func=unzip_fn)
                if self.sha1 is not None:
                    if os.path.split(download_loc)[-1] != self.sha1:
                        raise RuntimeError("The sha1 of the downloaded file does not match with the provided one")
                dcache.update({url: download_loc})
                write_json(dcache, os.path.join(self.data_download_cache, DATA_CACHE_CONF))
                return self._get_embedding_file(download_loc, self.embedding_key)




import torch
import torch.nn as nn
import numpy as np
from typing import Dict, List
from eight_mile.pytorch.layers import Dense, TransformerEncoderStack, TransformerEncoder, TransformerDecoderStack, TransformerDecoder, EmbeddingsStack
from eight_mile.pytorch.embeddings import LookupTableEmbeddings, LearnedPositionalLookupTableEmbeddingsWithBias

# BERT HuggingFace Tokenizers checkpoints can be converted into MEAD Baseline Transformer checkpoints
# With a simple name change
BERT_HF_LAYER_MAP = {
    ## FFN weights
    'bert.encoder.layer.{}.intermediate.dense.weight': 'generator.encoders.{}.ffn.0.layer.weight',
    'bert.encoder.layer.{}.intermediate.dense.bias': 'generator.encoders.{}.ffn.0.layer.bias',
    'bert.encoder.layer.{}.output.dense.weight': 'generator.encoders.{}.ffn.3.layer.weight',
    'bert.encoder.layer.{}.output.dense.bias': 'generator.encoders.{}.ffn.3.layer.bias',

    ## MHA weights
    'bert.encoder.layer.{}.attention.self.key.weight': 'generator.encoders.{}.self_attn.w_K.layer.weight',
    'bert.encoder.layer.{}.attention.self.key.bias': 'generator.encoders.{}.self_attn.w_K.layer.bias',
    'bert.encoder.layer.{}.attention.self.query.weight': 'generator.encoders.{}.self_attn.w_Q.layer.weight',
    'bert.encoder.layer.{}.attention.self.query.bias': 'generator.encoders.{}.self_attn.w_Q.layer.bias',
    'bert.encoder.layer.{}.attention.self.value.weight': 'generator.encoders.{}.self_attn.w_V.layer.weight',
    'bert.encoder.layer.{}.attention.self.value.bias': 'generator.encoders.{}.self_attn.w_V.layer.bias',
    'bert.encoder.layer.{}.attention.output.dense.weight': 'generator.encoders.{}.self_attn.w_O.layer.weight',
    'bert.encoder.layer.{}.attention.output.dense.bias': 'generator.encoders.{}.self_attn.w_O.layer.bias',

    ## LN weights
    # The names in of layer norm our transformers are a bit unspecific
    # think of ln1 as ln_x and ln2 as ln_attn_output
    'bert.encoder.layer.{}.output.LayerNorm.beta': 'generator.encoders.{}.ln1.bias',
    'bert.encoder.layer.{}.output.LayerNorm.gamma': 'generator.encoders.{}.ln1.weight',
    'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'generator.encoders.{}.ln2.bias',
    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'generator.encoders.{}.ln2.weight'
}

BERT_HF_FT_LAYER_MAP = {
    ## FFN weights
    'bert.encoder.layer.{}.intermediate.dense.weight': 'transformer.encoders.{}.ffn.0.layer.weight',
    'bert.encoder.layer.{}.intermediate.dense.bias': 'transformer.encoders.{}.ffn.0.layer.bias',
    'bert.encoder.layer.{}.output.dense.weight': 'transformer.encoders.{}.ffn.3.layer.weight',
    'bert.encoder.layer.{}.output.dense.bias': 'transformer.encoders.{}.ffn.3.layer.bias',

    ## MHA weights
    'bert.encoder.layer.{}.attention.self.key.weight': 'transformer.encoders.{}.self_attn.w_K.layer.weight',
    'bert.encoder.layer.{}.attention.self.key.bias': 'transformer.encoders.{}.self_attn.w_K.layer.bias',
    'bert.encoder.layer.{}.attention.self.query.weight': 'transformer.encoders.{}.self_attn.w_Q.layer.weight',
    'bert.encoder.layer.{}.attention.self.query.bias': 'transformer.encoders.{}.self_attn.w_Q.layer.bias',
    'bert.encoder.layer.{}.attention.self.value.weight': 'transformer.encoders.{}.self_attn.w_V.layer.weight',
    'bert.encoder.layer.{}.attention.self.value.bias': 'transformer.encoders.{}.self_attn.w_V.layer.bias',
    'bert.encoder.layer.{}.attention.output.dense.weight': 'transformer.encoders.{}.self_attn.w_O.layer.weight',
    'bert.encoder.layer.{}.attention.output.dense.bias': 'transformer.encoders.{}.self_attn.w_O.layer.bias',

    ## LN weights
    # The names in of layer norm our transformers are a bit unspecific
    # think of ln1 as ln_x and ln2 as ln_attn_output
    'bert.encoder.layer.{}.output.LayerNorm.beta': 'transformer.encoders.{}.ln1.bias',
    'bert.encoder.layer.{}.output.LayerNorm.gamma': 'transformer.encoders.{}.ln1.weight',
    'bert.encoder.layer.{}.attention.output.LayerNorm.beta': 'transformer.encoders.{}.ln2.bias',
    'bert.encoder.layer.{}.attention.output.LayerNorm.gamma': 'transformer.encoders.{}.ln2.weight'
}

BERT_HF_EMBED_MAP = {
    ## Embedding weights
    'bert.embeddings.word_embeddings.weight': 'embeddings.embeddings.0.embeddings.weight',
    'bert.embeddings.position_embeddings.weight': 'embeddings.embeddings.0.pos_embeddings.weight',
    'bert.embeddings.token_type_embeddings.weight': 'embeddings.embeddings.1.embeddings.weight',
    'bert.embeddings.LayerNorm.beta': 'embeddings.reduction.ln.bias',
    'bert.embeddings.LayerNorm.gamma': 'embeddings.reduction.ln.weight',
}


def convert_transformers_keys(num_layers: int, d: Dict, nested_layer_map: Dict = BERT_HF_LAYER_MAP, flat_map: Dict = BERT_HF_EMBED_MAP) -> Dict:
    m = {}
    for i in range(num_layers):
        for k, v in nested_layer_map.items():
            m[v.format(i)] = d[k.format(i)]

    for k, v in flat_map.items():
        m[v] = d[k]


    return m


def tlm_load_state_dict(module: nn.Module, checkpoint_file: str):
    """

    :param tlm: Safely loads the state dict for a transformer encoder
    :param checkpoint_file: The file name
    :return: None
    """
    from_str = 'transformer'
    to_str = 'generator'
    if hasattr(module, 'transformer'):
        from_str = 'generator'
        to_str = 'transformer'
    ckpt_dict = torch.load(checkpoint_file)
    renamed = {}
    for k, v in ckpt_dict.items():
        renamed[k.replace(from_str, to_str)] = v
    unmatch = module.load_state_dict(renamed, strict=False)
    if unmatch.missing_keys or len(unmatch.unexpected_keys) > 2:
        print("Warning: Embedding doesn't match with the checkpoint being loaded.")
        print(f"missing keys: {unmatch.missing_keys}\n unexpected keys: {unmatch.unexpected_keys}")


def to_weight_array(pytorch_layer: nn.Module, name: str) -> Dict:
    """Convert a {`LayerNorm`, `Linear`, `layers.Dense`} to `weights` and `bias` arrays

    :param pytorch_layer: A layer to get weights for
    :param name: The name of this layer to serialize
    :return: A Dictionary containing `weights` and `bias` keys
    """
    if isinstance(pytorch_layer, Dense):
        pytorch_layer = pytorch_layer.layer
    weights = pytorch_layer.weight.cpu().detach().numpy()
    bias = pytorch_layer.bias.cpu().detach().numpy()
    return {f"{name}/weights": weights, f"{name}/bias": bias}


def from_weight_array(pytorch_layer: nn.Module, d: Dict, name: str):
    """Read in {`LayerNorm`, `Linear`, `layers.Dense`} from `weights` and `bias` fields

    :param pytorch_layer: A layer to get weights for
    :param d: A Dict containing the arrays by key
    :param name: The name of this layer
    :return: None
    """
    if isinstance(pytorch_layer, Dense):
        pytorch_layer = pytorch_layer.layer
    device = pytorch_layer.weight.device
    pytorch_layer.weight = nn.Parameter(torch.from_numpy(d[f"{name}/weights"]).to(device=device), requires_grad=True)
    pytorch_layer.bias = nn.Parameter(torch.from_numpy(d[f"{name}/bias"]).to(device=device), requires_grad=True)


def to_ffn_array(pytorch_ffn: nn.Sequential, name: str) -> Dict:
    """Convert a `FFN` layer to a set of arrays

    :param pytorch_ffn: An `FFN` layer for Transformers
    :param name: The name of the layer
    :return: A Dict containing the arrays by key
    """
    d = {}
    d.update(to_weight_array(pytorch_ffn[0], f"{name}/expansion"))
    d.update(to_weight_array(pytorch_ffn[3], f"{name}/squeeze"))
    return d


def from_ffn_array(pytorch_ffn: nn.Sequential, d: Dict, name: str):
    """Restore an `FFN` layer's weights from a set of arrays

    :param pytorch_ffn: An `FFN` layer for Transformers
    :param d: A Dict containing the arrays by key
    :param name: The name of the layer
    :return: None
    """
    from_weight_array(pytorch_ffn[0], d, f"{name}/expansion")
    from_weight_array(pytorch_ffn[3], d, f"{name}/squeeze")


def to_attn_array(pytorch_attn: nn.Module, name: str) -> Dict:
    """Convert a self-attention module to a set of arrays

    :param pytorch_attn: The self-attention layer of the transformer encoder, could be MultiHeadedAttention or
    MultiHeadedRelativeAttention
    :param name: The name of the layer
    :return: A Dict containing the arrays by key
    """
    d = {}

    d.update(to_weight_array(pytorch_attn.w_Q, f"{name}/w_Q"))
    d.update(to_weight_array(pytorch_attn.w_K, f"{name}/w_K"))
    d.update(to_weight_array(pytorch_attn.w_V, f"{name}/w_V"))

    if hasattr(pytorch_attn, 'w_O'):
        d.update(to_weight_array(pytorch_attn.w_O, f"{name}/w_O"))

    if hasattr(pytorch_attn, 'rpr_key'):
        rpr_key_weights = pytorch_attn.rpr_key.weight.cpu().detach().numpy()
        d.update({f"{name}/rpr_key": rpr_key_weights})

    if hasattr(pytorch_attn, 'rpr_value'):
        rpr_value_weights = pytorch_attn.rpr_value.weight.cpu().detach().numpy()
        d.update({f"{name}/rpr_value": rpr_value_weights})

    return d


def from_attn_array(pytorch_attn: nn.Module, d: Dict, name: str):
    """Restore the self-attention module from a set of arrays

    :param pytorch_attn: A self-attention module, could be MultiHeadedAttention or MultiHeadedRelativeAttention
    :param d: A Dict of arrays by key
    :param name: The name of the layer
    """
    from_weight_array(pytorch_attn.w_Q, d, f"{name}/w_Q")
    from_weight_array(pytorch_attn.w_K, d, f"{name}/w_K")
    from_weight_array(pytorch_attn.w_V, d, f"{name}/w_V")

    if hasattr(pytorch_attn, 'w_O'):
        from_weight_array(pytorch_attn.w_O, d, f"{name}/w_O")

    if hasattr(pytorch_attn, 'rpr_key'):
        device = pytorch_attn.rpr_key.weight.device
        pytorch_attn.rpr_key.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/rpr_key"]).to(device=device))

    if hasattr(pytorch_attn, 'rpr_value'):
        device = pytorch_attn.rpr_key.weight.device
        pytorch_attn.rpr_value.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/rpr_value"]).to(device=device))


def to_encoder_array(pytorch_encoder: TransformerEncoder, name: str) -> Dict:
    """Convert a `TransformerEncoder` layer to an set of numpy arrays

    :param pytorch_encoder: A `TransformerEncoder` layer
    :param name: The layer name
    :return: A Dict of arrays by key
    """
    d = {}
    d.update(to_weight_array(pytorch_encoder.ln1, f"{name}/ln1"))
    d.update(to_weight_array(pytorch_encoder.ln2, f"{name}/ln2"))
    d.update(to_attn_array(pytorch_encoder.self_attn, f"{name}/attn"))
    d.update(to_ffn_array(pytorch_encoder.ffn, f"{name}/ffn"))
    return d


def from_encoder_array(pytorch_encoder: TransformerEncoder, d: Dict, name: str):
    """Restore a `TransformerEncoder` layer from a set of numpy arrays

    :param pytorch_encoder: A `TransformerEncoder` layer
    :param d: A Dict of arrays by key
    :param name: The layer name
    :return: None
    """
    from_weight_array(pytorch_encoder.ln1, d, f"{name}/ln1")
    from_weight_array(pytorch_encoder.ln2, d, f"{name}/ln2")
    from_attn_array(pytorch_encoder.self_attn, d, f"{name}/attn")
    from_ffn_array(pytorch_encoder.ffn, d, f"{name}/ffn")


def from_decoder_array(pytorch_decoder: TransformerDecoder, d: Dict, name: str):
    """Restore a `TransformerDecoder` layer from a set of numpy arrays

    :param pytorch_decoder: A `TransformerDecoder` layer
    :param d: A Dict of arrays by key
    :param name: The layer name
    :return: None
    """
    from_weight_array(pytorch_decoder.ln1, d, f"{name}/ln1")
    from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln2")
    from_weight_array(pytorch_decoder.ln2, d, f"{name}/ln3")
    from_attn_array(pytorch_decoder.src_attn, d, f"{name}/src_attn")
    from_attn_array(pytorch_decoder.self_attn, d, f"{name}/self_attn")
    from_ffn_array(pytorch_decoder.ffn, d, f"{name}/ffn")


def to_decoder_array(pytorch_decoder: TransformerDecoder, name: str) -> Dict:
    """Convert a `TransformerDeccoder` layer to an set of numpy arrays

    :param pytorch_decoder: A `TransformerDecoder` layer
    :param name: The layer name
    :return: A Dict of arrays by key
    """
    d = {}
    d.update(to_weight_array(pytorch_decoder.ln1, f"{name}/ln1"))
    d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln2"))
    d.update(to_weight_array(pytorch_decoder.ln2, f"{name}/ln3"))
    d.update(to_attn_array(pytorch_decoder.self_attn, f"{name}/self_attn"))
    d.update(to_attn_array(pytorch_decoder.src_attn, f"{name}/src_attn"))
    d.update(to_ffn_array(pytorch_decoder.ffn, f"{name}/ffn"))
    return d


def to_embed_array(pytorch_embed: nn.Module, name: str) -> Dict:
    """Convert positional embedding to a `weights` array, if it's learned positional embedding,
    save the pos_weight as well

    :param pytorch_embed: An embedding module
    :param name: A layer name
    :return: A Dict containing the embedding `weights`
    """
    d = {}
    weights = pytorch_embed.embeddings.weight.cpu().detach().numpy()
    d.update({f"{name}/weights": weights})
    if hasattr(pytorch_embed, 'pos_embeddings'):
        pos_weights = pytorch_embed.pos_embeddings.weight.cpu().detach().numpy()
        d.update({f"{name}/pos_weights": pos_weights})
    return d


def from_embed_array(pytorch_embed: nn.Module, d: Dict, name: str):
    """Restore a positional embedding from a `weights` array, if it's a learned positional embedding, the pos_weights
    is also restored

    :param pytorch_embed: An embedding module
    :param d: A Dict containing a `weights` array to restore
    :param name: name of the layer
    :return: None
    """
    device = pytorch_embed.embeddings.weight.device
    pytorch_embed.embeddings.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/weights"]).to(device=device), requires_grad=True)
    if hasattr(pytorch_embed, 'pos_embeddings'):
        pytorch_embed.pos_embeddings.weight = torch.nn.Parameter(torch.from_numpy(d[f"{name}/pos_weights"]).to(device=device),
                                                                 requires_grad=True)


def to_tlm_array(pytorch_tlm: nn.Module, embeddings_keys: List[str] = None, name: str = "TLM") -> Dict:
    """Convert a Transformer LM-type module to a set of weights in a Dict

    :param pytorch_tlm: A Transformer LM-type module
    :param embeddings_keys: A key to get the embeddings from, defaults to `None` in which case, gets all keys
    :param name: A name for this TLM
    :return: A Dict containing all the keys to restore from Embeddings and the TransformerEncoderStack
    """
    d = {}
    transformer = pytorch_tlm.transformer if hasattr(pytorch_tlm, 'transformer') else pytorch_tlm.generator
    d.update(to_encoder_stack_array(transformer, name=f"{name}/TransformerEncoderStack"))
    keys_to_write = embeddings_keys if embeddings_keys else list(pytorch_tlm.embeddings.keys())

    for embeddings_key in keys_to_write:
        d.update(to_embed_array(pytorch_tlm.embeddings[embeddings_key], name=f"{name}/Embeddings/{embeddings_key}"))

    if hasattr(pytorch_tlm.embeddings.reduction, 'ln'):
        d.update(to_weight_array(pytorch_tlm.embeddings.reduction.ln, name=f"{name}/Embeddings/reduction/ln"))
    return d


def save_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM", verbose: bool = False):
    """Save a TLM to an NPZ file

    :param pytorch_tlm: A Transformer LM-type module
    :param npz: A file to save
    :param embeddings_keys: A key to get embeddings from.  Defaults to `None`, in which case, all embeddings are written
    :param name: A name for this TLM
    :param verbose: whether output 
    :return: None
    """
    d = to_tlm_array(pytorch_tlm, embeddings_keys, name)
    if verbose:
        print(d.keys())
    np.savez(npz, **d)


def save_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
                                 tgt_embedding_key: str = 'y', name: str = "Seq2Seq", verbose: bool = False):
    """Save a Transformer seq2seq file out

    The will be in pytorch_seq2seq.encoder.transformer, and the usual conversions work for that (via `to_tlm_array()`).
    The decoder requires a new converter for the portion containing attention weights between the encoder and the decoder

    :param pytorch_seq2seq: A Transformer Seq2Seq module
    """
    #enc = to_tlm_array(tf_seq2seq.encoder, embeddings_keys, name=f'{name}/Encoder')

    enc = {}
    transformer = pytorch_seq2seq.encoder.transformer
    enc.update(to_encoder_stack_array(transformer, name=f"{name}/TransformerEncoderStack"))
    enc_keys_to_write = src_embeddings_keys if src_embeddings_keys else list(pytorch_seq2seq.src_embeddings.keys())

    for embeddings_key in enc_keys_to_write:
        enc.update(to_embed_array(pytorch_seq2seq.src_embeddings[embeddings_key], name=f"{name}/SrcEmbeddings/{embeddings_key}"))

    dec = {}
    transformer_decoder = pytorch_seq2seq.decoder.transformer_decoder

    dec.update(to_decoder_stack_array(transformer_decoder, name=f"{name}/TransformerDecoderStack"))
    dec.update(to_embed_array(pytorch_seq2seq.tgt_embedding, name=f"{name}/TgtEmbedding/{tgt_embedding_key}"))

    if verbose:
        print(enc.keys())
        print(dec.keys())
    np.savez(npz, **enc, **dec)


def to_encoder_stack_array(
    pytorch_encoder_stack: TransformerEncoderStack, name: str = "TransformerEncoderStack"
) -> Dict:
    """Convert a `TransformerEncoderStack` to a set of weigths

    :param pytorch_encoder_stack: A transformer encoder stack
    :param name: A name
    :return: A Dict containing a set of weights
    """
    d = {}
    if isinstance(pytorch_encoder_stack.ln, nn.LayerNorm):
        d.update(to_weight_array(pytorch_encoder_stack.ln, f"{name}/ln"))
    for i, enc_pyt in enumerate(pytorch_encoder_stack.encoders):
        d.update(to_encoder_array(enc_pyt, f"{name}/{i}"))
    return d


def from_encoder_stack_array(
    pytorch_encoder_stack: TransformerEncoderStack, d: Dict, name: str = "TransformerEncoderStack"
):
    """Restore weights from a `TransformerEncoderStack`

    :param pytorch_encoder_stack: A transformer encoder stack
    :param d: A Dict containing sets of arrays
    :param name: A name for this primitive
    :return: None
    """
    if isinstance(pytorch_encoder_stack.ln, nn.LayerNorm):
        from_weight_array(pytorch_encoder_stack.ln, d, f"{name}/ln")
    for i, enc_pyt in enumerate(pytorch_encoder_stack.encoders):
        from_encoder_array(enc_pyt, d, f"{name}/{i}")


def to_decoder_stack_array(
    pytorch_decoder_stack: TransformerDecoderStack, name: str = "TransformerDecoderStack"
) -> Dict:
    """Convert a `TransformerDecoderStack` to a set of weights

    :param pytorch_decoder_stack: A transformer decoder stack
    :param name: A name
    :return: A Dict containing a set of weights
    """
    d = {}
    if isinstance(pytorch_decoder_stack.ln, nn.LayerNorm):
        d.update(to_weight_array(pytorch_decoder_stack.ln, f"{name}/ln"))
    for i, dec_pytorch in enumerate(pytorch_decoder_stack.decoders):
        d.update(to_decoder_array(dec_pytorch, f"{name}/{i}"))
    return d


def from_decoder_stack_array(
    pytorch_decoder_stack: TransformerDecoderStack, d: Dict, name: str = "TransformerDecoderStack"
):
    """Restore weights from a `TransformerDecoderStack`

    :param pytorch_decoder_stack: A transformer decoder stack
    :param d: A Dict containing sets of arrays
    :param name: A name for this primitive
    :return: None
    """
    if isinstance(pytorch_decoder_stack.ln, nn.LayerNorm):
        from_weight_array(pytorch_decoder_stack.ln, d, f"{name}/ln")
    for i, dec_pyt in enumerate(pytorch_decoder_stack.decoders):
        from_decoder_array(dec_pyt, d, f"{name}/{i}")


def from_tlm_array(pytorch_tlm: nn.Module, d: Dict, embeddings_keys: List[str] = None, name: str = "TLM"):
    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning)

    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized

    :param pytorch_tlm: A TLM-like model
    :param d: A Dict of weights to restore for each layer
    :param embeddings_keys: Name of embeddings to restore, defaults to `None`, in which case all embeddings are restored
    :param name: A name for this primitive
    :return:
    """
    transformer = pytorch_tlm.transformer if hasattr(pytorch_tlm, 'transformer') else pytorch_tlm.generator
    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
    keys_to_restore = embeddings_keys if embeddings_keys else list(pytorch_tlm.embeddings.keys())
    for embeddings_key in keys_to_restore:
        from_embed_array(pytorch_tlm.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
        if isinstance(pytorch_tlm.embeddings[embeddings_key], LearnedPositionalLookupTableEmbeddingsWithBias):
            tt = LookupTableEmbeddings(vsz=2, dsz=pytorch_tlm.embeddings.output_dim)
            from_embed_array(tt, d, f"{name}/Embeddings/tt")
            pytorch_tlm.embeddings[embeddings_key].bias = nn.Parameter(tt.embeddings.weight[0])
        else:
            from_embed_array(pytorch_tlm.embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
    if hasattr(pytorch_tlm.embeddings.reduction, 'ln'):
        from_weight_array(pytorch_tlm.embeddings.reduction.ln, d, f"{name}/Embeddings/reduction/ln")


def load_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM"):
    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning

    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized

    :param pytorch_tlm: A TLM-like model
    :param npz: A file to restore the weights from
    :param embeddings_key: Name of embeddings to restore, defaults to `None` in which case we restore all embeddings
    :param name: A name for this primitive
    :return:
    """
    d = np.load(npz)
    from_tlm_array(pytorch_tlm, d, embeddings_keys, name)


def load_transformer_seq2seq_npz(pytorch_seq2seq: nn.Module, npz: str, src_embeddings_keys: List[str] = None,
                                 tgt_embedding_key: str = 'y', name: str = "Seq2Seq"):
    """Save a Transformer seq2seq file out

    The will be in pytorch_seq2seq.encoder.transformer, and the usual conversions work for that (via `to_tlm_array()`).
    The decoder requires a new converter for the portion containing attention weights between the encoder and the decoder

    :param pytorch_seq2seq: A Transformer Seq2Seq module
    :param npz: The file name
    :param src_embeddings_keys: An optional list of the src embeddings keys to load, otherwise use what we find
    :param tgt_embedding_key: An optional tgt embedding, otherwise assume 'y' (TODO: bad assumption?)
    :param name: An optional name of the model in the NPZ, otherwise assume `Seq2Seq`
    """

    d = np.load(npz)

    transformer = pytorch_seq2seq.encoder.transformer
    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")

    enc_keys_to_restore = src_embeddings_keys if src_embeddings_keys else list(pytorch_seq2seq.src_embeddings.keys())

    for embeddings_key in enc_keys_to_restore:
        from_embed_array(pytorch_seq2seq.src_embeddings[embeddings_key], d, f"{name}/SrcEmbeddings/{embeddings_key}")

    transformer_decoder = pytorch_seq2seq.decoder.transformer_decoder

    from_decoder_stack_array(transformer_decoder, d,  name=f"{name}/TransformerDecoderStack")
    from_embed_array(pytorch_seq2seq.decoder.tgt_embeddings, d, name=f"{name}/TgtEmbedding/{tgt_embedding_key}")



def seq2seq_enc_from_tlm_array(pytorch_tlm: nn.Module, d: Dict, embeddings_keys: List[str] = None, name: str = "TLM"):
    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning)
    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized
    :param pytorch_tlm: A TLM-like model
    :param d: A Dict of weights to restore for each layer
    :param embeddings_keys: Name of embeddings to restore, defaults to `None`, in which case all embeddings are restored
    :param name: A name for this primitive
    :return:
    """
    transformer = pytorch_tlm.encoder.transformer
    from_encoder_stack_array(transformer, d, name=f"{name}/TransformerEncoderStack")
    keys_to_restore = embeddings_keys if embeddings_keys else list(pytorch_tlm.src_embeddings.keys())

    for embeddings_key in keys_to_restore:
        from_embed_array(pytorch_tlm.src_embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
        if isinstance(pytorch_tlm.src_embeddings[embeddings_key], LearnedPositionalLookupTableEmbeddingsWithBias):
            tt = LookupTableEmbeddings(vsz=2, dsz=pytorch_tlm.embeddings.output_dim)
            from_embed_array(tt, d, f"{name}/Embeddings/tt")
            pytorch_tlm.src_embeddings[embeddings_key].bias = nn.Parameter(tt.embeddings.weight[0])
        else:
            from_embed_array(pytorch_tlm.src_embeddings[embeddings_key], d, f"{name}/Embeddings/{embeddings_key}")
    if hasattr(pytorch_tlm.src_embeddings.reduction, 'ln'):
        from_weight_array(pytorch_tlm.src_embeddings.reduction.ln, d, f"{name}/Embeddings/reduction/ln")


def load_seq2seq_enc_from_tlm_npz(pytorch_tlm: nn.Module, npz: str, embeddings_keys: List[str] = None, name: str = "TLM"):
    """Restore a TLM-like model (possibly a `nn.Module` for fine-tuning
    We just populate the `TransformerEncoderStack` and the embeddings from weights, all other values remain
    uninitialized
    :param pytorch_tlm: A TLM-like model
    :param npz: A file to restore the weights from
    :param embeddings_key: Name of embeddings to restore, defaults to `None` in which case we restore all embeddings
    :param name: A name for this primitive
    :return:
    """
    d = np.load(npz)
    seq2seq_enc_from_tlm_array(pytorch_tlm, d, embeddings_keys, name)

def load_tlm_transformers_bin(pytorch_tlm: nn.Module, bin_file: str, replace_layers=BERT_HF_LAYER_MAP, replace_embeds=BERT_HF_EMBED_MAP):
    """For BERT transformer from HuggingFace, we need a TLM with EmbeddingsStack with 2 features and LN reduce

    The Transformer architecture used by BERT mirrors T2T (with layer norms coming after each transformer layer and
    a layer norm immediately following a sum reduce of learned-positional embeddings, the word LUT embeddings and
    a token-type embedding.  For many cases, the token-type embeddings is uninteresting, and should be simply set
    to 0.  For some cases, setting the token-type is critical.  In MEAD, to support the token type we need a
    `LookupTableEmbeddings`, and for the token itself, we need a `LearnedPositionalEmbedding` (which adds the two
    features of the token LUT and the position LUT together.  MEAD composes multiple features with the
    `EmbeddingsStack` primitive, and provides a reduction operator.  This is equivalent to BERT if we supply
    the `EmbeddingsStack` with 2 features and the `sum-layer-norm` reduction.  To do this, 2 vectorizers must be supplied
    to MEAD, one for the usual wordpiece tokenizers, and another tokenizer which has to produce the exact same number
    of tokens but with a token type value.  For example, for sentence-based token type delimiting, we would want to
    have a vectorizer that starts with 0, and every time it sees a `[SEP]` in the token itself, produces an incremented
    value, e.g. 0.... 1..... 2....

    For applications where the token-type is uninteresting, we can actually make a more efficient representation,
    which we do by creating a single `LearnedPositionalLookupTableEmbeddingsWithBias` embedding object as the single
    feature to the `EmbeddingsStack`.  This operation looks like a `LearnedPositionalLookupTableEmbeddings` object
    but it has a learnable bias parameter which is initialized to LUT index 0 of the BERT pretrained checkpoint.
    Additionally, because this object's goal is to replace the compositional approach of having 2 features, it also has
    a LayerNorm operator under the hood.  If you use this object, under normal circumstances, you do not need to bother
    providing `EmbeddingsStack` with any reduction as there is only 1 embedding.

    We want to support both types of models being passed in here by the user.  The way we do this, is to check the
    number of features in the `EmbeddingsStack`.  If its just 1, we can just monkey-patch the object as though it was
    the 2-feature equivalent, and after its loaded, restore the single feature so that when it goes back to the user
    they have what the put in originally

    :param pytorch_tlm: A Transformer LM
    :param bin_file: A HuggingFace PyTorch BERT checkpoint
    :param replace_layers: The mapping from HuggingFace Transformer keys to MEAD Transformer keys
    :param replace_embeds: The mapping from HuggingFace Embeddings key to MEAD Embeddings Keys
    :return:
    """
    d = torch.load(bin_file)
    transformer = pytorch_tlm.transformer if hasattr(pytorch_tlm, 'transformer') else pytorch_tlm.generator
    num_layers = len(transformer.encoders)
    mapped_keys = convert_transformers_keys(num_layers, d, replace_layers, replace_embeds)
    old_embeddings_stack = None
    k_0 = pytorch_tlm.embeddings.keys()[0]

    # There are 2 options to consider here
    # Option 1: the user doesnt care about token type embeddings, which means that the embedding type will just be
    #   the usual LP embeddings with an added bias term set to weight 0
    # Option 2: the user does care about token types and has provided a token type feature (presumed to be in the
    #   second key of the embeddings stack

    if isinstance(pytorch_tlm.embeddings[k_0], LearnedPositionalLookupTableEmbeddingsWithBias):
        old_embeddings_stack = pytorch_tlm.embeddings
        # we need to temporarily monkey patch the embeddings to load them, and then we can reset them to what they were
        d = {k_0: pytorch_tlm.embeddings[k_0], 'tt':  LookupTableEmbeddings(vsz=2, dsz=old_embeddings_stack.output_dim)}
        pytorch_tlm.embeddings = EmbeddingsStack(d, reduction='sum-layer-norm')
    unknown_keys = pytorch_tlm.load_state_dict(mapped_keys, strict=False)
    missing_keys = [key for key in unknown_keys.missing_keys if key != 'embeddings.embeddings.0.bias']

    if old_embeddings_stack:
        old_embeddings_stack[k_0].bias = nn.Parameter(pytorch_tlm.embeddings['tt'].embeddings.weight[0])
        old_embeddings_stack.reduction.ln = pytorch_tlm.embeddings.reduction.ln
        pytorch_tlm.embeddings = old_embeddings_stack

    return {'missing': missing_keys, 'unexpected': unknown_keys.unexpected_keys}



from typing import Optional, Tuple
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure


__all__ = ["reliability_diagram", "reliability_curve"]


def reliability_diagram(
    accs: np.array,
    confs: np.array,
    left_bins: np.array,
    num_classes: Optional[int] = None,
    x_ticks: Optional[np.array] = None,
    y_ticks: Optional[np.array] = None,
    title: Optional[str] = "Reliability Graph",
    y_label: Optional[str] = "Accuracy",
    x_label: Optional[str] = "Confidence",
    gap_label: Optional[str] = "Gap",
    output_label: Optional[str] = "Outputs",
    gap_edge: str = "r",
    gap_color: str = "r",
    gap_alpha: float = 0.25,
    gap_hatch: str = "/",
    output_edge: str = "k",
    output_color: str = "b",
    fig_size: int = 3,
    ax: Optional[Axes] = None,
    fig: Optional[Figure] = None,
) -> Tuple[Figure, Axes]:
    """Plot a reliability graph a la https://arxiv.org/abs/1706.04599

    :param accs: The average accuracy for examples in that bin
    :param congs: The average confidence for examples in that bin
    :param left_bins: The bins that define the groups, they should specify the left
        edge of the bin
    :param x_ticks: Where to show the ticks along the x axis. If not provided it
        will display every other tick, pass an empty list to skip adding ticks
    :param y_ticks: Where to show the ticks along the y axis. If not provided it
        uses default matplotlib ticks, padd an empty list to skip adding ticks
    :param title: The title for the axes, pass None to skip adding a title to the axes
    :param y_label: A label for the y axes, pass None to skip adding a label
    :param x_label: A label for the x axes, pass None to skip adding a label
    :param gap_label: A label for the gap bins, pass None to skip adding a label
    :param output_label: A label for the output bins, pass None to skip adding a label
    :param gap_edge: The color to use for the edges of the gap bins
    :param gap_color: The color to use for the main section of the gap bins
    :param gap_hatch: The hatching pattern for the gap, useful when a bin a under confident
    :param output_edge: The color to use for the edges of the output bins
    :param output_color: The color to use for the main section of the output bins
    :param fig_size: The figure size, only used if creating the figure/axis from scratch
    :param ax: An Axes object that we can create the graph in, if None it will be created
    :param fig: A figure that holds the axes object

    :returns: `Tuple(fig, ax)` If the figure and ax are passed in they are returned, if they
        were not passed in the created ones are returned. The don't use the figure at all but
        because we want to return both in case we created them.
    """
    # If neither a figure non-axes are passed in we create them.
    if ax is None and fig is None:
        fig, ax = plt.subplots(1, 1, figsize=(fig_size, fig_size))
    # If the axes was not passed in but the figure was we get the current active
    # axes for the figure.
    if ax is None and fig is not None:
        ax = fig.gca()
    # If the axes is provided but the figure we look up the figure on the axes
    if ax is not None and fig is None:
        fig = ax.figure
    widths = np.diff(left_bins, append=1)
    expected = left_bins + (widths / 2)
    expected = np.where(confs != 0, confs, expected)
    if num_classes is not None:
        # This is the minimum possible confidence. This is maximum entropy. If your confidence is
        # lower than this and something else has to have a larger confidence
        min_conf = 1.0 / num_classes
        # These bins are lowest confidence allowed in the bin so the highest confidence is the left edge
        # of the bin above it. So if we do a compare to the bins we will see that the first bin will always
        # be false. But we can use a np.roll and force the last bin to be true to simulate as if we compared
        # to the right bin
        possible_mask = np.roll(left_bins > min_conf, shift=-1)
        possible_mask[-1] = True
        # Mask out the expected values for bins that we can't put values in. Otherwise we will see all these
        # red bars we can't do anything about.
        expected *= possible_mask
    ax.bar(
        left_bins,
        accs,
        align='edge',
        width=widths,
        label=output_label,
        edgecolor=output_edge,
        color=output_color
    )
    ax.bar(
        left_bins,
        expected - accs,
        bottom=accs,
        align='edge',
        width=widths,
        label=gap_label,
        edgecolor=gap_edge,
        color=gap_color,
        alpha=gap_alpha,
        hatch=gap_hatch,
    )
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c=".3")
    ax.set_aspect("equal", "box")
    ax.set_title(title)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if x_ticks is None:
        x_ticks = [y for i, y in enumerate(left_bins) if i % 2 == 0] + [1.0]
    ax.set_xticks(x_ticks)
    if y_ticks is not None:
        ax.set_yticks(y_ticks)
    if not (gap_label is None and output_label is None):
        ax.legend()
    return fig, ax


def reliability_curve(
    accs: np.array,
    confs: np.array,
    num_classes: Optional[int] = None,
    x_ticks: Optional[np.array] = None,
    y_ticks: Optional[np.array] = None,
    title: Optional[str] = "Reliability Curve",
    y_label: Optional[str] = "Accuracy",
    x_label: Optional[str] = "Confidence",
    label: Optional[str] = None,
    color: str = "tab:orange",
    line_style: str = "o-",
    fig_size: int = 3,
    ax: Optional[Axes] = None,
    fig: Optional[Figure] = None,
) -> Tuple[Figure, Axes]:
    """Plot a reliability curve a la https://scikit-learn.org/stable/modules/calibration.html#calibration

    Note:
        When there is a bin that had no samples in it the average confidence will be zero
        and the average accuracy will be zero too. If this is the case we skip plotting
        that point.

    :param confs: The average confidence in a bin
    :param accs: The average accuracy for examples in that bin
    :param x_ticks: Where to show the ticks along the x axis. If not provided it
        will display every other tick, pass an empty list to skip adding ticks
    :param y_ticks: Where to show the ticks along the y axis. If not provided it
        uses default matplotlib ticks, padd an empty list to skip adding ticks
    :param title: The title for the axes, pass None to skip adding a title to the axes
    :param y_label: A label for the y axes, pass None to skip adding a label
    :param x_label: A label for the x axes, pass None to skip adding a label
    :param label: A label for the plot we are making.
    :param color: A color for the line we are making.
    :param line_style: The stle of line to draw.
    :param fig_size: The figure size, only used if creating the figure/axis from scratch
    :param ax: An Axes object that we can create the graph in, if None it will be created
    :param fig: A figure that holds the axes object

    :returns: `Tuple(fig, ax)` If the figure and ax are passed in they are returned, if they
        were not passed in the created ones are returned. The don't use the figure at all but
        because we want to return both in case we created them.
    """
    # If neither a figure non-axes are passed in we create them.
    if ax is None and fig is None:
        fig, ax = plt.subplots(1, 1, figsize=(fig_size, fig_size))
    # If the axes was not passed in but the figure was we get the current active
    # axes for the figure.
    if ax is None and fig is not None:
        ax = fig.gca()
    # If the axes is provided but the figure we look up the figure on the axes
    if ax is not None and fig is None:
        fig = ax.figure
    valid = (confs != 0) | (accs != 0)
    ax.plot(
        confs[valid], accs[valid], line_style, color=color, label=label
    )
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c=".3")
    ax.set_aspect("equal", "box")
    ax.set_title(title)
    ax.set_ylabel(y_label)
    ax.set_xlabel(x_label)
    if x_ticks is None:
        x_ticks = [y for i, y in enumerate(np.arange(0, 1 + 1e-8, 1. / len(confs))) if i % 2 == 0]
    ax.set_xticks(x_ticks)
    if y_ticks is not None:
        ax.set_yticks(y_ticks)
    if label is not None:
        ax.legend()
    return fig, ax


def _demo():
    f, axs = plt.subplots(2, 2, figsize=(4, 4))
    data = np.array([0, 0, .15, .25, .28, .40, .38, .42, .56, .84])
    x = np.arange(0, 1, 1 / len(data))
    reliability_diagram(data, x + 0.05, x, num_classes=10, ax=axs[0][0], title="Reliability Diagram (10 classes)")
    data[5] = 0
    reliability_diagram(data, x + 0.05, x, num_classes=100, ax=axs[0][1], title="Reliability Diagram (100 classes)")
    conf = np.array([
        0.07167232, 0.16942227, 0.26184   , 0.35503329, 0.44979852,
        0.55041263, 0.64713617, 0.73971958, 0.83297067, 0.92842073
    ])
    acc = np.array([
        0.        , 0.00589623, 0.00789084, 0.03013831, 0.18707902,
        0.64843534, 0.92417131, 0.98329403, 0.98977505, 1.
    ])
    reliability_curve(acc, conf, label="SVC", ax=axs[1][0])
    conf = np.array([
        0.07167232, 0.16942227, 0.26184   , 0.35503329, 0.44979852,
        0.55041263, 0.64713617, 0, 0.83297067, 0.92842073
    ])
    acc = np.array([
        0.        , 0.00589623, 0.00789084, 0.03013831, 0.18707902,
        0.64843534, 0.92417131, 0, 0.98977505, 1.
    ])
    reliability_curve(acc * 0.5, conf, label="NB", ax=axs[1][1], color="b")
    reliability_curve(acc, conf, label="SVC", ax=axs[1][1], title="Reliability Curve, Missing bins")
    plt.show()


if __name__ == "__main__":
    _demo()



import math
import logging
import torch
import torch.autograd
from eight_mile.optz import create_lr_scheduler, register_lr_scheduler
from eight_mile.optz import (
    ConstantScheduler,
    WarmupLinearScheduler,
    CyclicLRScheduler,
    PiecewiseDecayScheduler,
    ZarembaDecayScheduler,
    CosineDecayScheduler,
    InverseTimeDecayScheduler,
    ExponentialDecayScheduler,
    CompositeLRScheduler,
)

logger = logging.getLogger("mead.layers")


@register_lr_scheduler(name="default")
class ConstantSchedulerPyTorch(ConstantScheduler):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@register_lr_scheduler(name="warmup_linear")
class WarmupLinearSchedulerPyTorch(WarmupLinearScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="clr")
class CyclicLRSchedulerPyTorch(CyclicLRScheduler):
    def __init__(self, *args, **kwargs):
        super().__init(*args, **kwargs)


@register_lr_scheduler(name="piecewise")
class PiecewiseDecaySchedulerPyTorch(PiecewiseDecayScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="zaremba")
class ZarembaDecaySchedulerPyTorch(ZarembaDecayScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="cosine")
class CosineDecaySchedulerPyTorch(CosineDecayScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="invtime")
class InverseTimeDecaySchedulerPytorch(InverseTimeDecayScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="exponential")
class ExponentialDecaySchedulerPyTorch(ExponentialDecayScheduler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@register_lr_scheduler(name="composite")
class CompositeLRSchedulerPyTorch(CompositeLRScheduler):
    pass


class AdamW(torch.optim.Optimizer):
    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1.0 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
                denom = exp_avg_sq.sqrt().add_(group["eps"])

                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1
                p.data.addcdiv_(exp_avg, denom, value=-step_size)

                if group["weight_decay"] != 0.0:
                    p.data.add_(p.data, alpha=-group["weight_decay"] * group["lr"])

        return loss


class OptimizerManager:
    def __init__(self, model_or_params, global_step=0, **kwargs):
        if isinstance(model_or_params, torch.nn.Module):
            parameters = model_or_params.parameters()
        else:
            parameters = model_or_params
        self.global_step = global_step
        if "lr_function" in kwargs:
            self.lr_function = kwargs["lr_function"]
        else:
            if "lr_scheduler_type" not in kwargs:
                kwargs["lr_scheduler_type"] = "default"
            self.lr_function = create_lr_scheduler(**kwargs)
        self._init_optimizer(parameters, **kwargs)
        self.current_lr = 0

    @property
    def global_step(self):
        return self._global_step

    @global_step.setter
    def global_step(self, value):
        self._global_step = value

    def _init_optimizer(self, parameters, **kwargs):
        wd = float(kwargs.get("weight_decay", 0))
        optim = kwargs.get("optim", "sgd")
        self.current_lr = kwargs.get("eta", kwargs.get("lr", 0.01))
        if optim == "adadelta":
            logger.info("adadelta(eta=%f, wd=%f)", self.current_lr, wd)
            self.optimizer = torch.optim.Adadelta(parameters, lr=self.current_lr, weight_decay=wd)
        elif optim.startswith("adam"):
            beta1 = kwargs.get("beta1", 0.9)
            beta2 = kwargs.get("beta2", 0.999)
            eps = kwargs.get("epsilon", 1e-8)
            if optim == "adam":
                logger.info(
                    "adam(eta=%f, beta1=%f, beta2=%f, epsilon=%f, wd=%f)", self.current_lr, beta1, beta2, eps, wd
                )
                self.optimizer = torch.optim.Adam(
                    parameters, lr=self.current_lr, betas=(beta1, beta2), eps=eps, weight_decay=wd
                )
            elif optim == "adamw":
                logger.info(
                    "adamw(eta=%f, beta1=%f, beta2=%f, epsilon=%f, wd=%f)", self.current_lr, beta1, beta2, eps, wd
                )
                self.optimizer = AdamW(
                    parameters,
                    lr=self.current_lr,
                    betas=(beta1, beta2),
                    eps=eps,
                    weight_decay=wd,
                )
        elif optim == "rmsprop":
            mom = kwargs.get("mom", 0.0)
            logger.info("rmsprop(eta=%f, wd=%f, mom=%f)", self.current_lr, wd, mom)
            self.optimizer = torch.optim.RMSprop(parameters, lr=self.current_lr, weight_decay=wd, momentum=mom)
        elif optim == "asgd":
            logger.info("asgd(eta=%f, wd=%f)", self.current_lr, wd)
            self.optimizer = torch.optim.ASGD(parameters, lr=self.current_lr, weight_decay=wd)
        else:
            mom = kwargs.get("mom", 0.9)
            logger.info("sgd(eta=%f, mom=%f, wd=%f)", self.current_lr, mom, wd)
            self.optimizer = torch.optim.SGD(parameters, lr=self.current_lr, momentum=mom, weight_decay=wd)

    def _identity(self, _):
        return self.current_lr

    def step(self):
        """Runs at every step and updates the learning rate

        :return:
        """
        self.optimizer.step()
        self.current_lr = self.update_lr()
        self.global_step += 1

    def zero_grad(self):
        self.optimizer.zero_grad()

    def update_lr(self):
        lr = self.lr_function(self.global_step)
        for p in self.optimizer.param_groups:
            p["lr"] = lr

        return lr


class EagerOptimizer:
    def __init__(self, loss, optimizer=None, **kwargs):
        self.loss = loss
        if optimizer:
            self.optimizer = optimizer
        else:
            self.optimizer_args = kwargs
            self.optimizer = None

    def update(self, model, x, y):
        if not self.optimizer:
            self.optimizer = OptimizerManager(model, **self.optimizer_args)
        self.optimizer.zero_grad()
        l = self.loss(model, x, y)
        l.backward()
        self.optimizer.step()
        return float(l)

    def update_with_hidden(self, model, h, x, y):
        if not self.optimizer:
            self.optimizer = OptimizerManager(model, **self.optimizer_args)
        self.optimizer.zero_grad()
        l, h = self.loss(model, h, x, y)
        l.backward()
        self.optimizer.step()
        return float(l), h



import io
import copy
import logging
import collections
import contextlib
import numpy as np
from eight_mile.utils import optional_params, exporter, write_json, read_config_file, Offsets, mime_type, revlut
__all__ = []
export = exporter(__all__)
logger = logging.getLogger("mead.layers")


@export
def norm_weights(word_vectors):
    norms = np.linalg.norm(word_vectors, axis=1, keepdims=True)
    norms = (norms == 0) + norms
    return word_vectors / norms


@export
def write_word2vec_file(filename, vocab, word_vectors):
    """Write out a binary word2vec file

    This function allows either a list of vocabulary of the same size
    as the vectors, in which case it does a simple write, or a vocabulary
    dict which may contain "magic" Offset values that baseline uses.  If
    those are populated offsets in the dict, they will occupy the low indices
    (first values), so to prune them, remove them from the vocab and shift the
    indices down

    :param filename: (`str`) The filename
    :param vocab: (`Dict[str, int]` or `List[str]`) A list of vocabulary
    :param word_vectors: The word vectors to write for each vocab
    :return: None
    """

    offset = 0
    # Check for case where we have a dict and possibly magic values to prune
    if isinstance(vocab, collections.Mapping):
        vocab_copy = copy.deepcopy(vocab)

        for v in Offsets.VALUES:
            if v in vocab_copy:
                del vocab_copy[v]
                offset += 1
        vocab_list = [0] * len(vocab_copy)
        for word, idx in vocab_copy.items():
            vocab_list[idx - offset] = word
    # Otherwise its just a list dont do anything weird
    else:
        vocab_list = vocab

    # Writing the file is pretty simple, just need the vocab size and depth
    # Then write each line
    with io.open(filename, "wb") as f:
        word_vectors_offset = word_vectors[offset:]
        vsz = len(vocab_list)
        dsz = word_vectors[0].shape[0]
        f.write(bytes("{} {}\n".format(vsz, dsz), encoding="utf-8"))
        assert len(vocab_list) == len(word_vectors_offset)
        for word, vector in zip(vocab_list, word_vectors_offset):
            vec_str = vector.tobytes()
            f.write(bytes("{} ".format(word), encoding="utf-8") + vec_str)


@export
class EmbeddingsModel(object):
    def __init__(self):
        super().__init__()

    def get_dsz(self):
        pass

    def get_vsz(self):
        pass

    def get_vocab(self):
        pass

    def save_md(self, target):
        pass


def pool_vec(embeddings, tokens, operation=np.mean):
    if type(tokens) is str:
        tokens = tokens.split()
    try:
        return operation([embeddings.lookup(t, False) for t in tokens], 0)
    except:
        return embeddings.weights[0]


@export
class WordEmbeddingsModel(EmbeddingsModel):
    def __init__(self, **kwargs):
        super().__init__()
        self.vocab = kwargs.get("vocab")
        self.vsz = kwargs.get("vsz")
        self.dsz = kwargs.get("dsz")
        self.weights = kwargs.get("weights")
        if "md_file" in kwargs:
            md = read_config_file(kwargs["md_file"])
            self.vocab = md["vocab"]
            self.vsz = md["vsz"]
            self.dsz = md["dsz"]
        if "weights_file" in kwargs:
            self.weights = np.load(kwargs["weights_file"]).get("arr_0")

        if self.weights is not None:
            if self.vsz is None:
                self.vsz = self.weights.shape[0]
            else:
                assert self.vsz == self.weights.shape[0]
            if self.dsz is None:
                self.dsz = self.weights.shape[1]
            else:
                assert self.dsz == self.weights.shape[1]

        elif self.vsz is not None and self.dsz is not None:
            self.weights = np.zeros((self.vsz, self.dsz))

    def get_dsz(self):
        return self.dsz

    def get_vsz(self):
        return self.vsz

    def get_vocab(self):
        return self.vocab

    def get_weights(self):
        return self.weights

    def save_md(self, target):
        write_json({"vsz": self.get_vsz(), "dsz": self.get_dsz(), "vocab": self.get_vocab()}, target)

    def save_weights(self, target):
        np.savez(target, self.weights)

    def lookup(self, word, nullifabsent=True):
        if word in self.vocab:
            return self.weights[self.vocab[word]]
        if nullifabsent:
            return None
        return self.nullv

    def __getitem__(self, word):
        return self.lookup(word, nullifabsent=False)


@export
class PretrainedEmbeddingsModel(WordEmbeddingsModel):
    def __init__(self, filename, known_vocab=None, unif_weight=None, keep_unused=False, normalize=False, **kwargs):
        super().__init__()

        if (known_vocab is None or not known_vocab) and keep_unused is False:
            logger.warning(
                "Warning: known_vocab=None or is Empty, keep_unused=False. Setting keep_unused=True, all vocab will be preserved"
            )
            keep_unused = True
        uw = 0.0 if unif_weight is None else unif_weight
        self.vocab = {}
        # Set the start offset to one past the last special token
        idx = Offsets.OFFSET

        word_vectors, self.dsz, known_vocab, idx = self._read_vectors(filename, idx, known_vocab, keep_unused, **kwargs)
        self.nullv = np.zeros(self.dsz, dtype=np.float32)
        special_tokens = [self.nullv]
        for i in range(1, len(Offsets.VALUES)):
            special_tokens.append(np.random.uniform(-uw, uw, self.dsz).astype(np.float32))
        word_vectors = special_tokens + word_vectors
        # Add "well-known" values to the vocab
        for i, name in enumerate(Offsets.VALUES):
            self.vocab[name] = i

        if known_vocab is not None:
            # Remove "well-known" values
            for name in Offsets.VALUES:
                known_vocab.pop(name, 0)
            unknown = {v: cnt for v, cnt in known_vocab.items() if cnt > 0}
            for v in unknown:
                word_vectors.append(np.random.uniform(-uw, uw, self.dsz).astype(np.float32))
                self.vocab[v] = idx
                idx += 1

        self.weights = np.array(word_vectors)
        if normalize is True:
            self.weights = norm_weights(self.weights)

        self.vsz = self.weights.shape[0]
        assert self.weights.dtype == np.float32

    def _read_vectors(self, filename, idx, known_vocab, keep_unused, **kwargs):
        use_mmap = bool(kwargs.get("use_mmap", False))
        read_fn = self._read_word2vec_file
        is_glove_file = mime_type(filename) == "text/plain"
        if use_mmap:
            if is_glove_file:
                read_fn = self._read_text_mmap
            else:
                read_fn = self._read_word2vec_mmap
        elif is_glove_file:
            read_fn = self._read_text_file

        return read_fn(filename, idx, known_vocab, keep_unused)

    def _read_word2vec_file(self, filename, idx, known_vocab, keep_unused):
        word_vectors = []
        with io.open(filename, "rb") as f:
            header = f.readline()
            vsz, dsz = map(int, header.split())
            width = 4 * dsz
            for i in range(vsz):
                word = self._readtospc(f)
                raw = f.read(width)
                if word in self.vocab:
                    continue
                if keep_unused is False and word not in known_vocab:
                    continue
                if known_vocab and word in known_vocab:
                    known_vocab[word] = 0
                vec = np.fromstring(raw, dtype=np.float32)
                word_vectors.append(vec)
                self.vocab[word] = idx
                idx += 1
        return word_vectors, dsz, known_vocab, idx

    @staticmethod
    def _read_word2vec_line_mmap(m, width, start):
        current = start + 1
        while m[current : current + 1] != b" ":
            current += 1
        vocab = m[start:current].decode("utf-8").strip(" \n")
        raw = m[current + 1 : current + width + 1]
        value = np.fromstring(raw, dtype=np.float32)
        return vocab, value, current + width + 1

    def _read_word2vec_mmap(self, filename, idx, known_vocab, keep_unused):
        import mmap

        word_vectors = []
        with io.open(filename, "rb") as f:
            with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
                header_end = m[:50].find(b"\n")
                vsz, dsz = map(int, (m[:header_end]).split(b" "))
                width = 4 * dsz
                current = header_end + 1
                for i in range(vsz):
                    word, vec, current = self._read_word2vec_line_mmap(m, width, current)
                    if word in self.vocab:
                        continue
                    if keep_unused is False and word not in known_vocab:
                        continue
                    if known_vocab and word in known_vocab:
                        known_vocab[word] = 0

                    word_vectors.append(vec)
                    self.vocab[word] = idx
                    idx += 1
                return word_vectors, dsz, known_vocab, idx

    @staticmethod
    def _readtospc(f):

        s = bytearray()
        ch = f.read(1)

        while ch != b" ":
            s.extend(ch)
            ch = f.read(1)
        s = s.decode("utf-8")
        # Only strip out normal space and \n not other spaces which are words.
        return s.strip(" \n")

    def _read_text_file(self, filename, idx, known_vocab, keep_unused):
        word_vectors = []

        with io.open(filename, "r", encoding="utf-8") as f:
            for i, line in enumerate(f):
                line = line.rstrip("\n ")
                values = line.split(" ")
                word = values[0]
                if i == 0 and len(values) == 2:
                    print("VSZ: {}, DSZ: {}".format(word, values[1]))
                    continue
                if word in self.vocab:
                    continue
                if keep_unused is False and word not in known_vocab or word in self.vocab:
                    continue
                if known_vocab and word in known_vocab:
                    known_vocab[word] = 0
                vec = np.asarray(values[1:], dtype=np.float32)
                word_vectors.append(vec)
                self.vocab[word] = idx
                idx += 1
        dsz = vec.shape[0]
        return word_vectors, dsz, known_vocab, idx

    def _read_text_mmap(self, filename, idx, known_vocab, keep_unused):
        import mmap

        word_vectors = []
        with io.open(filename, "r", encoding="utf-8") as f:
            with contextlib.closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as m:
                for i, line in enumerate(iter(m.readline, b"")):
                    line = line.rstrip(b"\n")
                    values = line.split(b" ")
                    if i == 0 and len(values) == 2:
                        print("VSZ: {}, DSZ: {}".format(values[0], values[1]))
                        continue
                    if len(values) == 0:
                        break
                    word = values[0].decode("utf-8").strip(" \n")
                    if word in self.vocab:
                        continue
                    if keep_unused is False and word not in known_vocab:
                        continue
                    if known_vocab and word in known_vocab:
                        known_vocab[word] = 0
                    vec = np.asarray(values[1:], dtype=np.float32)
                    word_vectors.append(vec)
                    self.vocab[word] = idx
                    idx += 1
                dsz = vec.shape[0]
                return word_vectors, dsz, known_vocab, idx


@export
class PretrainedEmbeddingsStack(EmbeddingsModel):
    def __init__(self, filenames, known_vocab, counts=True, unif_weight=None, normalize=False, **kwargs):
        uw = 0.0 if unif_weight is None else unif_weight

        self.vocab = dict()
        for i, name in enumerate(Offsets.VALUES):
            self.vocab[name] = i
        self.vsz = Offsets.OFFSET

        if counts is True:
            for name in Offsets.VALUES:
                known_vocab.pop(name, 0)
            attested = [v for v, cnt in known_vocab.items() if cnt > 0]
            for k, v in enumerate(attested):
                self.vocab[v] = k + Offsets.OFFSET
                self.vsz += 1
        else:
            self.vocab = known_vocab
            self.vsz = max(self.vocab.values()) + 1

        index2word = revlut(self.vocab)
        # vocab = word2index
        embeddings = []

        for file in filenames:
            embeddings.append(PretrainedEmbeddingsModel(file, known_vocab))

        self.dsz = sum([embedding.dsz for embedding in embeddings])
        self.weights = np.random.uniform(-uw, uw, (self.vsz, self.dsz)).astype(np.float32)

        for i in range(len(self.vocab.keys())):
                w = index2word[i]
                e = []
                for emb in embeddings:
                    e.append(emb.lookup(w, False))
                self.weights[i] = np.concatenate(e)
        if normalize is True:
            self.weights = norm_weights(self.weights)

    def __getitem__(self, word):
        return self.lookup(word, nullifabsent=False)

    def lookup(self, word, nullifabsent=True):
        if word in self.vocab:
            return self.weights[self.vocab[word]]
        if nullifabsent:
            return None
        return np.zeros(self.dsz, dtype=np.float32)

    def get_vocab(self):
        return self.vocab

    def get_dsz(self):
        return self.dsz

    def get_vsz(self):
        return self.vsz

    def save_md(self, target):
        write_json({"vsz": self.get_vsz(), "dsz": self.get_dsz(), "vocab": self.get_vocab()}, target)



@export
class RandomInitVecModel(EmbeddingsModel):
    def __init__(self, dsz, known_vocab, counts=True, unif_weight=None):
        super().__init__()
        uw = 0.0 if unif_weight is None else unif_weight
        self.vocab = dict()
        for i, name in enumerate(Offsets.VALUES):
            self.vocab[name] = i
        self.dsz = dsz
        self.vsz = Offsets.OFFSET

        if counts is True:
            for name in Offsets.VALUES:
                known_vocab.pop(name, 0)
            attested = [v for v, cnt in known_vocab.items() if cnt > 0]
            for k, v in enumerate(attested):
                self.vocab[v] = k + Offsets.OFFSET
                self.vsz += 1
        else:
            self.vocab = known_vocab
            self.vsz = max(self.vocab.values()) + 1

        self.weights = np.random.uniform(-uw, uw, (self.vsz, self.dsz)).astype(np.float32)

        self.nullv = np.zeros(self.dsz, dtype=np.float32)

        self.weights[0] = self.nullv
        for i in range(1, len(Offsets.VALUES)):
            self.weights[i] = np.random.uniform(-uw, uw, self.dsz).astype(np.float32)

    def __getitem__(self, word):
        return self.lookup(word, nullifabsent=False)

    def lookup(self, word, nullifabsent=True):
        if word in self.vocab:
            return self.weights[self.vocab[word]]
        if nullifabsent:
            return None
        return self.nullv

    def get_vocab(self):
        return self.vocab

    def get_dsz(self):
        return self.dsz

    def get_vsz(self):
        return self.vsz

    def save_md(self, target):
        write_json({"vsz": self.get_vsz(), "dsz": self.get_dsz(), "vocab": self.get_vocab()}, target)






import math
import copy
import logging
import numpy as np
import tensorflow as tf
from eight_mile.utils import write_json, Offsets, is_sequence, calc_nfeats
from eight_mile.tf.layers import *


FLOAT32 = 4
GB2 = 1024 * 1024 * 1024 * 2
logger = logging.getLogger("baseline")


class TensorFlowEmbeddings(tf.keras.layers.Layer):
    """This provides a base for TensorFlow embeddings sub-graphs

    """

    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        # tf.kers.layers.Layer has a validation step that only allows certain kwargs
        # to be passed into it. These are not documented and you need to look into the
        # code to find this. For now just don't pass in out kwargs
        super().__init__(trainable=trainable, name=name, dtype=dtype)
        self.W = None

    def get_dsz(self):
        """Get the number of output dimension of this operation

        :return:
        """
        pass

    def get_vsz(self):
        """Get the number of words (including <PAD>) in the vocabulary

        :return:
        """
        pass

    def get_weights(self):
        raise NotImplementedError

    def encode(self, *x):
        """This defines the computation of the sub-graph for this object and returns the output node

        :return:
        """
        pass

    @property
    def output_dim(self):
        return self.get_dsz()

    def call(self, x):
        return self.encode(x)

    def get_feed_dict(self):
        """Return a feed dict that is needed to initialize this embeddings."""
        return {}


class LookupTableEmbeddings(TensorFlowEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, cpu_placement=False, **kwargs):
        """Create a lookup-table based embedding.

        :param name: The name of the feature/placeholder, and a key for the scope
        :param kwargs:

        :Keyword Arguments: See below
        * *vsz* -- (``int``) this is the vocabulary (input) size of the lookup table
        * *dsz* -- (``int``) the output dimension size of this embedding
        * *finetune* -- (``bool``) (default is `True`) should we allow the sub-graph to learn updated weights
        * *weights* -- (``numpy.ndarray``) Optional `vsz x dsz` weight matrix for initialization
        * *unif* -- (``float``) (defaults to `0.1`) If the weights should be created, what is the random initialization range
        """
        trainable = kwargs.get("finetune", trainable)
        # The layers have a filter of allowed keywords and the docs don't list what they are
        # you need to look in code. We are just not passing kwargs for now.
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.vsz = kwargs.get("vsz")
        self.padding_idx = kwargs.get('padding_idx', Offsets.PAD)
        self.dsz = kwargs.get("dsz")
        self.finetune = kwargs.get("finetune", trainable)
        self.scope = kwargs.get("scope", "LUT")
        self.dropin = kwargs.get("dropin", 0.0)
        self._weights = kwargs.get("weights")
        self.drop = tf.keras.layers.Dropout(rate=self.dropin, noise_shape=(self.get_vsz(), 1))
        self.cpu_placement = cpu_placement
        if self._weights is None:
            unif = kwargs.get("unif", 0.1)
            self._weights = np.random.uniform(-unif, unif, (self.vsz, self.dsz))
        else:
            self.vsz, self.dsz = self._weights.shape

    def build(self, input_shape):

        if self.cpu_placement:
            with tf.device("cpu:0"):
                self.W = self.add_weight(
                        name=f"{self.scope}/Weight",
                        shape=(self.vsz, self.dsz),
                        initializer=tf.constant_initializer(self._weights),
                        trainable=self.finetune,
                )
        else:
            self.W = self.add_weight(
                name=f"{self.scope}/Weight",
                shape=(self.vsz, self.dsz),
                initializer=tf.constant_initializer(self._weights),
                trainable=self.finetune,
            )
        super().build(input_shape)

    def _embed_w_dropout(self, x):
        # The ablation table (4) in https://arxiv.org/pdf/1708.02182.pdf shows this has a massive impact
        embedding_w_dropout = self.drop(self.W, training=TRAIN_FLAG())
        word_embeddings = tf.nn.embedding_lookup(embedding_w_dropout, x)
        return word_embeddings

    def encode(self, *x):
        """Build a simple Lookup Table and set as input `x` if it exists, or `self.x` otherwise.

        :param x: An optional input sub-graph to bind to this operation or use `self.x` if `None`
        :return: The sub-graph output
        """
        self.x = x[0]
        if self.padding_idx is not None:
            e0 = tf.tensor_scatter_nd_update(
                self.W, tf.constant(self.padding_idx, dtype=tf.int32, shape=[1, 1]), tf.zeros(shape=[1, self.dsz])
            )
            with tf.control_dependencies([e0]):
                return self._embed_w_dropout(self.x)
        else:
            return self._embed_w_dropout(self.x)

    def get_vsz(self):
        return self.vsz

    def get_dsz(self):
        return self.dsz

    def get_weights(self):
        return self.W


class CharConvEmbeddings(TensorFlowEmbeddings):
    """dos Santos embeddings extended to parallel filters (AKA Kim character-aware neural language model inputs)

    """

    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.cpu_placement = bool(kwargs.get('cpu_placement', False))
        self.scope = kwargs.get("scope", "CharConv")
        self.finetune = kwargs.get("finetune", trainable)
        self.nfeat_factor = kwargs.get("nfeat_factor", None)
        self.cfiltsz = kwargs.get("cfiltsz", kwargs.get("filtsz", [3]))
        self.max_feat = kwargs.get("max_feat", 30)
        gating = kwargs.get("gating", "skip")
        num_gates = kwargs.get("num_gates", 1)
        activation = kwargs.get("activation", "tanh")
        self.wsz = kwargs.get("wsz", 30)
        self.projsz = kwargs.get("projsz")
        self.x = None
        # These are the actual final filter sizes and num features
        filtsz, nfeats = calc_nfeats(self.cfiltsz, self.nfeat_factor, self.max_feat, self.wsz)

        self.embed = LookupTableEmbeddings(name=f"{self.name}/CharLUT", finetune=self.finetune, **kwargs)
        dsz = self.embed.output_dim
        self.parallel_conv = ParallelConv(dsz, nfeats, filtsz, activation)
        self.gating_fns = tf.keras.Sequential()
        for _ in range(num_gates):
            if gating == 'skip':
                self.gating_fns.add(SkipConnection(self.parallel_conv.output_dim, activation))
            else:
                self.gating_fns.add(Highway(self.parallel_conv.output_dim))

        self.outsz = self.parallel_conv.output_dim
        if self.projsz is not None:
            self.outsz = self.projsz
            self.proj = tf.keras.layers.Dense(self.outsz, bias_initializer=tf.constant_initializer(0.0))



    @property
    def dsz(self):
        return self.outsz

    def encode(self, *x):
        self.x = x[0]

        mxlen = tf.shape(self.x)[1]
        mxwlen = tf.shape(self.x)[-1]
        char_bt_x_w = tf.reshape(self.x, [-1, mxwlen])
        cembed = self.embed(char_bt_x_w)
        cmot = self.parallel_conv(cembed)
        cmot = self.gating_fns(cmot)
        if self.projsz:
            cmot = self.proj(cmot)
        word_char = tf.reshape(cmot, [-1, mxlen, self.outsz])
        return word_char

    def get_vsz(self):
        return self.embed.get_vsz()

    def get_dsz(self):
        return self.outsz


class CharLSTMEmbeddings(TensorFlowEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.scope = kwargs.get("scope", "CharLUT")
        self.finetune = kwargs.get("finetune", trainable)
        self.lstmsz = kwargs.get("lstmsz", 50)
        self.lstm_layers = kwargs.get("layers", 1)
        self.pdrop = kwargs.get("pdrop", 0.5)
        self.rnn_type = kwargs.get("rnn_type", "blstm")
        self.x = None
        self.embed = LookupTableEmbeddings(name=f"{self.name}/CharLUT", finetune=self.finetune, **kwargs)
        self.lstm = BiLSTMEncoderHidden(
            self.embed.output_dim,
            self.lstmsz,
            self.lstm_layers,
            pdrop=self.pdrop,
            requires_length=True,
            name=f"{self.name}/blstm",
        )

    def encode(self, *x):
        self.x = x[0]
        shape = tf.shape(self.x)
        B = shape[0]
        T = shape[1]
        W = shape[2]
        flat_chars = tf.reshape(x, [-1, W])
        embed_chars = self.embed(flat_chars)

        # Calculate the lengths of each word
        word_lengths = tf.reduce_sum(tf.cast(tf.not_equal(flat_chars, Offsets.PAD), tf.int32), axis=1)

        # cuDNN throws an error if there is an input with a length of 0, this happens when the "word"
        # is actually a "<PAD>" so there are no characters to run the LSTM over. Here we just say
        # that the lengths is 1. This will make cudnn happy and we will just get junk in that spot
        patched_lengths = tf.math.maximum(word_lengths, 1)

        # Run the LSTM
        result = self.lstm((embed_chars, patched_lengths))

        # Create a mask that is true when the length is 0 (where the word was a pad) so that
        # we can mask out the junk that the lstm created because we needed a length of 1
        result = tf.multiply(result, tf.expand_dims(tf.cast(tf.not_equal(word_lengths, 0), tf.float32), -1))

        return tf.reshape(result, (B, T, self.lstmsz))

    def call(self, inputs):
        return self.encode(inputs)

    def get_dsz(self):
        return self.lstmsz

    def get_vsz(self):
        return self.embed.get_vsz()


class CharTransformerEmbeddings(TensorFlowEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.scope = kwargs.get("scope", "CharLUT")
        self.finetune = kwargs.get("finetune", trainable)
        self.embed = LookupTableEmbeddings(name=f"{self.name}/CharLUT", finetune=self.finetune, **kwargs)
        self.d_model = kwargs.get("wsz", 30)
        self.num_heads = kwargs.get("num_heads", 3)
        self.rpr_k = kwargs.get("rpr_k", 10)
        layers = kwargs.get("layers", 1)
        pdrop = kwargs.get("pdrop", 0.5)
        self.char_comp = TransformerEncoderStackWithLengths(
            self.num_heads,
            self.d_model,
            pdrop,
            False,
            layers,
            rpr_k=self.rpr_k,
            input_sz=self.embed.output_dim,
            name=f"{self.name}/transformer",
        )

    def encode(self, *x):
        self.x = x[0]
        shape = tf.shape(self.x)
        B = shape[0]
        T = shape[1]
        W = shape[2]
        flat_chars = tf.reshape(x, [-1, W])
        embed_chars = self.embed(flat_chars)

        # Calculate the lengths of each word
        lengths = tf.reduce_sum(tf.cast(tf.not_equal(flat_chars, Offsets.PAD), tf.int32), axis=1)

        # Run the LSTM
        result = self.char_comp((embed_chars, lengths))

        pooled = tf.reduce_max(result, -2, keepdims=False)

        return tf.reshape(pooled, (B, T, self.d_model))

    def call(self, inputs):
        return self.encode(inputs)

    def get_dsz(self):
        return self.d_model

    def get_vsz(self):
        return self.embed.get_vsz()


class PositionalMixin(tf.keras.layers.Layer):
    def positional(self, length):
        pass


class SinusoidalPositionalMixin(PositionalMixin):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)

        max_timescale = kwargs.get("max_timescale", 1.0e4)
        # Match the mxlen pytorch has because it precomputes the timing signal
        mxlen = kwargs.get('mxlen', 10000)

        word_dsz = self.get_dsz()
        log_timescale_increment = math.log(max_timescale) / float(word_dsz)
        inv_timescales = np.exp(np.arange(0, word_dsz, 2, dtype=np.float32) * -log_timescale_increment)

        pe = np.zeros((mxlen, word_dsz), dtype=np.float32)
        position = np.expand_dims(np.arange(0, mxlen, dtype=np.float32), 1)

        pe[:, 0::2] = np.sin(position * inv_timescales)
        pe[:, 1::2] = np.cos(position * inv_timescales)

        self.pe = tf.expand_dims(pe, 0)

    def positional(self, length):
        return self.pe[:, :length]


class LearnedPositionalMixin(PositionalMixin):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.mxlen = int(kwargs.get("mxlen", 512))
        self.pos_weights = kwargs.get("pos_weights")
        if self.pos_weights is None:
            unif = float(kwargs.get("unif", 0.1))
            self.pos_weights = np.random.uniform(-unif, unif, (self.mxlen, self.get_dsz()))

    def build(self, input_shape):
        self.pos = self.add_weight(
            name="pos",
            initializer=tf.constant_initializer(self.pos_weights),
            shape=[self.mxlen, self.get_dsz()],
            trainable=self.finetune,
        )
        super().build(input_shape)

    def positional(self, length):
        return tf.expand_dims(tf.nn.embedding_lookup(self.pos, tf.range(length, dtype=tf.int32)), 0)


class PositionalLookupTableEmbeddings(SinusoidalPositionalMixin, LookupTableEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(name=name, **kwargs)
        self.scale = math.sqrt(self.get_dsz())
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x) * tf.constant(self.scale)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())


class LearnedPositionalLookupTableEmbeddings(LearnedPositionalMixin, LookupTableEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(name=name, **kwargs)
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())


class LearnedPositionalLookupTableEmbeddingsWithBias(LearnedPositionalMixin, LookupTableEmbeddings):
    """Learned positional lookup table embeddings wih a bias and layer norm

    This is just a typical learned positional embedding but with a learnable
    bias and a layer norm.  This is equivalent to BERT embeddings when the
    token_type is not set

    """
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(name=name, **kwargs)

    def build(self, input_shape):
        super().build(input_shape)
        self.bias = self.add_weight(
            name="bias",
            initializer=tf.constant_initializer(0.0),
            shape=[1, self.get_dsz()],
            trainable=self.finetune,
        )

    def encode(self, *x):
        y = super().encode(*x)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        y = y + pos + self.bias
        return y


class PositionalCharConvEmbeddings(SinusoidalPositionalMixin, CharConvEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(name=name, **kwargs)
        self.scale = math.sqrt(self.get_dsz())
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x) * tf.constant(self.scale)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())


class LearnedPositionalCharConvEmbeddings(LearnedPositionalMixin, CharConvEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        super().__init__(name=name, **kwargs)
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())


class PositionalCharLSTMEmbeddings(SinusoidalPositionalMixin, CharLSTMEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.scale = math.sqrt(self.get_dsz())
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x) * tf.constant(self.scale)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())


class LearnedPositionalCharLSTMEmbeddings(LearnedPositionalMixin, CharLSTMEmbeddings):
    def __init__(self, trainable=True, name=None, dtype=tf.float32, **kwargs):
        trainable = kwargs.get("finetune", trainable)
        super().__init__(trainable=trainable, name=name, dtype=dtype, **kwargs)
        self.dropout = tf.keras.layers.Dropout(kwargs.get("dropout", 0.0))

    def encode(self, *x):
        y = super().encode(*x)
        T = tf.shape(y)[1]
        pos = self.positional(T)
        return self.dropout(y + pos, training=TRAIN_FLAG())



from typing import Optional
from collections import namedtuple
import numpy as np

__all__ = ["calibration_bins", "multiclass_calibration_bins", "binary_calibration_bins", "average_confidence", "Bins"]


Bins = namedtuple("Bins", "accs confs counts edges")


def multiclass_calibration_bins(truth: np.ndarray, probs: np.ndarray, bins: int, class_weights: Optional[np.ndarray] = None) -> Bins:
    """Calculate the binned confidence and accuracy for a multiclass problem.

    :param truth: A 1D array of the true labels for some examples.
    :param probs: A 1D array of the probabilities from a model. Each row represents an example in the dataset.
        and each column represents the probably assigned by the model to each class for that example.
    :param bins: The number of bins to use when aggregating.
    :param class_weights: A 1D array of scores that can add extra weight to examples of specific classes.

    :returns: The metrics aggregated by bins
    """
    preds = np.argmax(probs, axis=1)
    pred_probs = np.max(probs, axis=1)
    credit = truth == preds
    if class_weights is not None:
        weighted_labels = class_weights[truth]
        credit = credit * weighted_labels
    return binary_calibration_bins(credit, pred_probs, bins=bins)


calibration_bins = multiclass_calibration_bins


def binary_calibration_bins(credit: np.ndarray, probs: np.ndarray, bins: int) -> Bins:
    """Calculate the accuracy and confidence inside bins. This is similar to the sklearn function.

    Note:
        This is different than the multiclass function. This will look at the scores for a specific
        class even if it is below the 1/num_classes threshold that will cause the multiclass
        version to not select this class.

    :param credit: How much credit the model gets for this example. If it is wrong the value will be zero
        If it is right one. You can also give different classes different weights by passing real values.
    :param probs: The probabilities assigned to the positive class by the model.
    :param bins: The number of bins to use when aggregating.

    :returns: The metrics aggregated by bins
    """
    bins = np.linspace(0.0, 1.0, num=bins + 1, endpoint=True)

    bin_idx = np.digitize(probs, bins) - 1

    bin_conf_sum = np.bincount(bin_idx, weights=probs, minlength=len(bins))
    bin_acc_sum = np.bincount(bin_idx, weights=credit, minlength=len(bins))
    bin_counts = np.bincount(bin_idx, minlength=len(bins))

    mask = bin_counts == 0
    denom = bin_counts + mask

    bin_mean_conf = bin_conf_sum / denom
    bin_mean_acc = bin_acc_sum / denom

    return Bins(bin_mean_acc[:-1], bin_mean_conf[:-1], bin_counts[:-1], bins[:-1])


def average_confidence(probs: np.ndarray) -> float:
    """Calculate the average (maximum) confidence for a collection of predictions

    :param probs: `[B, C]` A matrix of probabilities, each row is an example and
        each column is a class.
    """
    if probs.ndim == 1:
        probs = np.expand_dims(probs, axis=-1)
    return np.mean(np.max(probs, axis=1))
